
/*
 * routines to collect dynamic data about shift instructions
 *
 * Copyright (C) 1999   Rick Gorton gorton@world.std.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */



#include <stdio.h>
#include "datatypes.h"

extern char *Inst_Logfile_Name;
/*
 * Data declarations local to this file
 */
long	Total_Procs;
long	Total_Params;
PROC_TABLE_TYPE		*dyn_proc_count;
PARAM_TABLE_TYPE	*dyn_param_table;
long	total_proc_count;
long	total_hits;
long	dyn_total_params;

	/*
	 * Constants for guessing which routines might be good
	 * candidates for glacial variable specialization
	 * HOT_PROC_TRIGGER_LEVEL - threshold for a
	 *	particular procedure to be a good candidate (1%)
	 * GLACIAL_PARAM_TRIGGER_LEVEL - PERCENT threshold for a
	 *	particular parameter to be a good candidate
	 * GLACIAL_DELTA_TRIGGER_LEVEL - PERCENT threshold for a
	 *	particular delta parameter to be a good candidate
	 */

#define	HOT_PROC_TRIGGER_LEVEL		0.01
#define	GLACIAL_PARAM_TRIGGER_LEVEL	20.0
#define	GLACIAL_DELTA_TRIGGER_LEVEL	5.0

void Dump_Procedure_Info ();

/*
 * Initialization
 */
void DynamicApplicationInit()
{
  int	i;

  for (i=0; i<= (int)LAST_OP; i++)
    {
    dyn_count[i] = 0x0;
    }
}

/*
 * Print out results
 */
void DumpDynamicData()
{
  int   i;
  long	netval;
  double	tmpval;

  fprintf (Inst_Logfile, "logfile, static proc, dyn proc, dyn params, dyn param matches (size %d)\n",
	PARAM_TABLE_TYPE_SIZE);
  fprintf (Inst_Logfile, "%s, %ld, %ld, %ld, %ld\n",
			Inst_Logfile_Name,
			Total_Procs, total_proc_count,
			dyn_total_params, total_hits);
  fprintf(Inst_Logfile, "*** Dynamic data end ***\n");
}


/*
 * Procedure data collection
 */

void Dyn_ProcCountAllocate (long elts)
{
  Total_Procs = elts;
  dyn_proc_count = (PROC_TABLE_TYPE *) calloc (sizeof(PROC_TABLE_TYPE), elts);
  total_proc_count = 0;
}

void Dyn_ProcCountInit (long PSeq, char *Pname,
				long A0, long A1, long A2,
				long FP0, long FP1, long FP2)
{
  strncat (dyn_proc_count[PSeq].ProcName, Pname, PROC_NAME_SIZE);
  dyn_proc_count[PSeq].A0_SeqNum = A0;
  dyn_proc_count[PSeq].A1_SeqNum = A1;
  dyn_proc_count[PSeq].A2_SeqNum = A2;
  dyn_proc_count[PSeq].FP0_SeqNum = FP0;
  dyn_proc_count[PSeq].FP1_SeqNum = FP1;
  dyn_proc_count[PSeq].FP2_SeqNum = FP2;
  dyn_proc_count[PSeq].counter = 0x0;
}

void Dyn_ParamCountAllocate (long elts)
{
  Total_Params = elts;
  dyn_param_table = (PARAM_TABLE_TYPE *) calloc (sizeof(PARAM_TABLE_TYPE), elts);
}

void Dyn_ParamCountInit ()
{
  /*
   * Used for resetting/zeroing the parameter tables (possibly during a run)
   */
  long	i, j;

  for (i = 0; i < Total_Params; i++)
    {
    for (j = 0; j < PARAM_TABLE_TYPE_SIZE; j++)
	{
	dyn_param_table[i].data_val[j].longstuff = 0x0;
	dyn_param_table[i].hits[j] = 0x0;
	}
    dyn_param_table[i].delta_from_prior.longstuff = 0x0;
    dyn_param_table[i].delta_hits = 0x0;
    dyn_param_table[i].delta_restarts = 0x0;
    }
  total_hits = 0x0;
  dyn_total_params = 0x0;
}

void Procedure_Count (long ProcIndex)
{
  long	i;

  dyn_proc_count[ProcIndex].counter++;
  total_proc_count++;

	/*
	 * Pseudo-temporal triggered events.
	 * Specifically for POV-Ray right now.
	 * The idea is to perform data dumps (and reset the tables)
	 * for a couple of different portions of the match rendering.
	 */
  return;
#ifdef RCGHACK
#endif
  if ((total_proc_count == 6000000)
	|| (total_proc_count == 30000000))
    {
    fprintf (Inst_Logfile,
		"Dynamic Data dump and reset @ %ld procedures starting\n",
		total_proc_count);
    for (i = 0; i < Total_Procs; i++)
	Dump_Procedure_Info (i);
    Dyn_ParamCountInit ();
    fprintf (Inst_Logfile,
		"Dynamic Data dump and reset @ %ld procedures completed\n",
		total_proc_count);
    }
}

void Glissade_Int (long	SeqIndex, long OperandValue)
{
  int	i;
  long	this_delta;

  dyn_total_params++;
  for (i = 0; i < PARAM_TABLE_TYPE_SIZE; i++)
    {
    if (dyn_param_table[SeqIndex].data_val[i].longstuff == OperandValue)
	{
	dyn_param_table[SeqIndex].hits[i]++;
	total_hits++;
	return;
	}
    }

	/* Miss - check for a constant delta hit */
  this_delta = OperandValue - dyn_param_table[SeqIndex].data_val[0].longstuff;
  if (this_delta == dyn_param_table[SeqIndex].delta_from_prior.longstuff)
    {

#ifdef DEBUG
fprintf(Inst_Logfile,
		"DELTA HIT (%d): value: 0x%lx\tPrior Delta 0x%lx\tThis Delta 0x%lx\n",
		SeqIndex,
		OperandValue,
		dyn_param_table[SeqIndex].delta_from_prior.longstuff,
		this_delta);
#endif

    dyn_param_table[SeqIndex].delta_hits++;
    dyn_param_table[SeqIndex].data_val[0].longstuff = OperandValue;
    dyn_param_table[SeqIndex].hits[0]++;
    return;
    }
  else
    {
    dyn_param_table[SeqIndex].delta_from_prior.longstuff = this_delta;
    dyn_param_table[SeqIndex].delta_hits = 0x0;
    dyn_param_table[SeqIndex].delta_restarts++;
    }

	/* Age contents */
  for (i = PARAM_TABLE_TYPE_SIZE - 1; i > 0; i--)
    {
    dyn_param_table[SeqIndex].data_val[i].longstuff =
		dyn_param_table[SeqIndex].data_val[i-1].longstuff;
    dyn_param_table[SeqIndex].hits[i] =
		dyn_param_table[SeqIndex].hits[i-1];
    }
	/* Insert new */
  dyn_param_table[SeqIndex].data_val[0].longstuff = OperandValue;
  dyn_param_table[SeqIndex].hits[0] = 0;
}

void Glissade_Float (long SeqIndex, double OperandValue)
{
  int	i;
  IEEE_UNION_	this_temp;
  double	this_delta;
  BOOLEAN	is_normalized;

	/*
	 * Evaluate FP values in integer form.  printf() seems to
	 * save f17, but the value in f17 seems to be a [random]
	 * denormalized number under some circumstances.  Comparing
	 * for equality against denormalized values causes a trap.
	 * Speed is important, so we are _not_ going to build this
	 * with any sort of software completion stuff.
	 */

  dyn_total_params++;
  this_temp.doublestuff = OperandValue;
  for (i = 0; i < PARAM_TABLE_TYPE_SIZE; i++)
    {
    if (dyn_param_table[SeqIndex].data_val[i].longstuff == this_temp.longstuff)
	{
	dyn_param_table[SeqIndex].hits[i]++;
	total_hits++;
	return;
	}
    }

	/*
	 * Miss - check for a constant delta hit but _only_ for normalized
	 * IEEE operand values.
	 * Treat a match as a match against slot zero.
	 */
  is_normalized = TRUE;

	/* check new value */
  if ((this_temp.rawbits.exponent == 0x0) && (this_temp.rawbits.mantissa != 0x0))
    is_normalized = FALSE;	/* Denormalized */
  if (this_temp.rawbits.exponent == 0x7ff)
    is_normalized = FALSE;	/* NaN or +/- Infinity */

	/* check current table value */
  if ((dyn_param_table[SeqIndex].data_val[0].rawbits.exponent == 0x0)
	&& (dyn_param_table[SeqIndex].data_val[0].rawbits.mantissa != 0x0))
    is_normalized = FALSE;	/* Denormalized */
  if (dyn_param_table[SeqIndex].data_val[0].rawbits.exponent == 0x7ff)
    is_normalized = FALSE;	/* NaN or +/- Infinity */

  if (is_normalized)
    {
    this_delta = OperandValue - dyn_param_table[SeqIndex].data_val[0].doublestuff;
    if (this_delta == dyn_param_table[SeqIndex].delta_from_prior.doublestuff)
	{
	dyn_param_table[SeqIndex].delta_hits++;
	dyn_param_table[SeqIndex].data_val[0].doublestuff = OperandValue;
	dyn_param_table[SeqIndex].hits[0]++;
	return;
	}
    else
	{
	dyn_param_table[SeqIndex].delta_from_prior.doublestuff = this_delta;
	dyn_param_table[SeqIndex].delta_hits = 0x0;
	dyn_param_table[SeqIndex].delta_restarts++;
	}
    }

	/* Age contents */
  for (i = PARAM_TABLE_TYPE_SIZE - 1; i > 0; i--)
    {
    dyn_param_table[SeqIndex].data_val[i].doublestuff =
		dyn_param_table[SeqIndex].data_val[i-1].doublestuff;
    dyn_param_table[SeqIndex].hits[i] =
		dyn_param_table[SeqIndex].hits[i-1];
    }
	/* Insert new */
  dyn_param_table[SeqIndex].data_val[0].doublestuff = OperandValue;
  dyn_param_table[SeqIndex].hits[0] = 0;
}

void Dump_Param_Info(char *namestring, long this_SeqNum,
			char *this_ProcName, long Proc_SeqIndex,
			BOOLEAN	Is_A_Float)
{
  long	i;
  double	proc_rate;
  double	param_rate;
  double	delta_rate;
  BOOLEAN	delta_seen;

  proc_rate = (double) dyn_proc_count[Proc_SeqIndex].counter / (double) total_proc_count;
  delta_rate = 100.0e0 * (double) dyn_param_table[this_SeqNum].delta_hits
			/ (double) dyn_proc_count[Proc_SeqIndex].counter;
#ifdef DEBUG
  fprintf(Inst_Logfile, "\t%s (%ld):\n", namestring, this_SeqNum);
#else
  fprintf(Inst_Logfile, "\t%s:\n", namestring);
#endif
  if ((dyn_param_table[this_SeqNum].delta_hits > 0)
		&& (delta_rate > GLACIAL_DELTA_TRIGGER_LEVEL)
		&& (proc_rate >= HOT_PROC_TRIGGER_LEVEL))
    {
    delta_seen = TRUE;
    if (Is_A_Float)
	fprintf(Inst_Logfile,
		"\tDELTA PARAMETER %s: %s %ld hits; %ld restarts (Delta currently %f)\n",
			this_ProcName, namestring,
			dyn_param_table[this_SeqNum].delta_hits,
			dyn_param_table[this_SeqNum].delta_restarts,
			dyn_param_table[this_SeqNum].delta_from_prior.doublestuff);
    else
	fprintf(Inst_Logfile,
		"\tDELTA PARAMETER %s: %s %ld hits; %ld restarts (Delta == 0x%lx)\n",
			this_ProcName, namestring,
			dyn_param_table[this_SeqNum].delta_hits,
			dyn_param_table[this_SeqNum].delta_restarts,
			dyn_param_table[this_SeqNum].delta_from_prior.longstuff);
    fprintf(Inst_Logfile, "\t\tNOTE: Slot 0 value may be final value\n");
    }
  else
    delta_seen = FALSE;
  for (i = 0; i < PARAM_TABLE_TYPE_SIZE; i++)
    {
    param_rate = 100.0e0 * (double) dyn_param_table[this_SeqNum].hits[i]
			/ (double) dyn_proc_count[Proc_SeqIndex].counter;
    if ((param_rate > GLACIAL_PARAM_TRIGGER_LEVEL) &&
		 (proc_rate >= HOT_PROC_TRIGGER_LEVEL))
	{
	char	tmpbuf[64];

#ifdef DEBUG
	fprintf(Inst_Logfile, "\tGLACIAL PARAMETER %s: %s (%ld) 0x%lx\n",
			this_ProcName, namestring, this_SeqNum,
			dyn_param_table[this_SeqNum].data_val[i].longstuff);
#else
	if (Is_A_Float)
	    {
	    if (dyn_param_table[this_SeqNum].hits[i] >= dyn_proc_count[Proc_SeqIndex].counter - 1)
		/* 100% - permit off by one due to table being initialized to zero */
		fprintf(Inst_Logfile, "\tCONSTANT PARAMETER %s: %s %f\n",
			this_ProcName, namestring,
			dyn_param_table[this_SeqNum].data_val[i].doublestuff);
	    else
		fprintf(Inst_Logfile, "\tGLACIAL PARAMETER %s: %s %f\n",
			this_ProcName, namestring,
			dyn_param_table[this_SeqNum].data_val[i].doublestuff);
	    }
	else
	    {
	    if (dyn_param_table[this_SeqNum].hits[i] >= dyn_proc_count[Proc_SeqIndex].counter - 1)
		/* 100% - permit off by one due to table being initialized to zero */
		fprintf(Inst_Logfile, "\tCONSTANT PARAMETER %s: %s 0x%lx\n",
			this_ProcName, namestring,
			dyn_param_table[this_SeqNum].data_val[i].longstuff);
	    else
		fprintf(Inst_Logfile, "\tGLACIAL PARAMETER %s: %s 0x%lx\n",
			this_ProcName, namestring,
			dyn_param_table[this_SeqNum].data_val[i].longstuff);
	    }
#endif
	}
    if ((dyn_param_table[this_SeqNum].hits[i] > 0)
		|| (delta_seen && (i == 0)))
	{
	fprintf(Inst_Logfile, "\t\tSlot %ld hits: %ld == (%2.2f%%)\tvalue:\t",
		i,
		dyn_param_table[this_SeqNum].hits[i],
		param_rate);
	if (Is_A_Float)
	    fprintf(Inst_Logfile, "%f\n",
		dyn_param_table[this_SeqNum].data_val[i].doublestuff);
	else
	    fprintf(Inst_Logfile, "0x%lx\n",
		dyn_param_table[this_SeqNum].data_val[i].longstuff);
	}
    }
}

void Dump_Procedure_Info (long SeqIndex)
{
#ifdef SUMMARY_ONLY
  return;
#endif

  if (dyn_proc_count[SeqIndex].counter < 2)
    return;
  fprintf(Inst_Logfile, "%s: called %ld times\n",
			dyn_proc_count[SeqIndex].ProcName,
                        dyn_proc_count[SeqIndex].counter);

  if (dyn_proc_count[SeqIndex].A0_SeqNum != -1)
    Dump_Param_Info ("A0", dyn_proc_count[SeqIndex].A0_SeqNum,
			dyn_proc_count[SeqIndex].ProcName, SeqIndex, FALSE);
  if (dyn_proc_count[SeqIndex].A1_SeqNum != -1)
    Dump_Param_Info ("A1", dyn_proc_count[SeqIndex].A1_SeqNum,
			dyn_proc_count[SeqIndex].ProcName, SeqIndex, FALSE);
  if (dyn_proc_count[SeqIndex].A2_SeqNum != -1)
    Dump_Param_Info ("A2", dyn_proc_count[SeqIndex].A2_SeqNum,
			dyn_proc_count[SeqIndex].ProcName, SeqIndex, FALSE);

  if (dyn_proc_count[SeqIndex].FP0_SeqNum != -1)
    Dump_Param_Info ("FP0", dyn_proc_count[SeqIndex].FP0_SeqNum,
			dyn_proc_count[SeqIndex].ProcName, SeqIndex, TRUE);
  if (dyn_proc_count[SeqIndex].FP1_SeqNum != -1)
    Dump_Param_Info ("FP1", dyn_proc_count[SeqIndex].FP1_SeqNum,
			dyn_proc_count[SeqIndex].ProcName, SeqIndex, TRUE);
  if (dyn_proc_count[SeqIndex].FP2_SeqNum != -1)
    Dump_Param_Info ("FP2", dyn_proc_count[SeqIndex].FP2_SeqNum,
			dyn_proc_count[SeqIndex].ProcName, SeqIndex, TRUE);
}
