
/*
 *	Atom program to hunt for, and report about Glacial variables on a
 *	routine basis.  We'll make some guesses as to what the parameter
 *	registers are based upon said registers being used before they are
 *	killed within the first Basic Block.  In order to try to more
 *	reasonably bound the problem set on this first attempt, I'll only
 *	look at the first two integer and fp parameters (if any).
 *
 * Copyright (C) 1999   Rick Gorton gorton@world.std.com
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */



#include <stdio.h>
#include <cmplrs/atom.inst.h>
#include "datatypes.h"
#include "declare.h"

#define	PARAM_REGISTER_MASK_	0x001f0000001f0000
#define	PARAM_A0_MASK_		0x10000
#define	PARAM_A1_MASK_		0x20000
#define	PARAM_A2_MASK_		0x40000
#define	PARAM_FP0_MASK_		0x0001000000000000
#define	PARAM_FP1_MASK_		0x0002000000000000
#define	PARAM_FP2_MASK_		0x0004000000000000

	/*
	 * Integer registers 16-21
	 * Float registers 16-21
	 */

void InstrumentInit(int iargc, char **iargv)
{
  int	i,j;
  char	namebuffer[256];

  if (iargc > 1)
    strcpy(namebuffer, iargv[1]);
  else
    strcpy(namebuffer, LOGFILE_NAME_);
  Inst_Logfile = fopen(namebuffer, "a+");
  printf("Logfile will be '%s'\n", namebuffer);
  fprintf(Inst_Logfile, "Logfile:\t%s\n", namebuffer);
  fprintf (Inst_Logfile, "*** Static data begin ***\n");

  ProcSequenceNum = 0;
  ParamSequenceNum = 0;

  AddCallProto ("Procedure_Count(long)");
  AddCallProto ("Glissade_Int(long, REGV)");
  AddCallProto ("Glissade_Float(long, FREGV)");
  AddCallProto ("Dump_Procedure_Info(long)");

  AddCallProto("DynamicInit(char *)");
  AddCallProgram(ProgramBefore, "DynamicInit", namebuffer);
}

Instrument(int argc, char **argv, Obj *obj)
{
  Proc *p; Block *b;  Inst *this_inst;
  RegvType	RegA, RegB, RegC;
  long		this_PC;

  fprintf(Inst_Logfile, "Object:\t%s\n", GetObjName(obj));
  fprintf(Inst_Logfile, "Number of Procedures:\t%d\n", GetObjInfo(obj, ObjNumberProcs));

  for (p = GetFirstObjProc(obj); p != NULL; p = GetNextProc(p))
    {
    InstRegUsageVec	this_proc_vec;
    long		param_vec;
    long		A0_SeqNum;
    long		A1_SeqNum;
    long		A2_SeqNum;
    long		FP0_SeqNum;
    long		FP1_SeqNum;
    long		FP2_SeqNum;

	/* Only first block relevant here */
#ifndef SUMMARY_ONLY
    fprintf(Inst_Logfile, "%s\t@ 0x%lx\t: ", ProcName(p), ProcPC(p));
#endif
    param_vec = 0x0;
    this_proc_vec.dreg_bitvec[0] = 0x0;

    for (b = GetFirstBlock(p); b != NULL; b = GetNextBlock(b))
	{
	/*
	 * Note - we don't care about pc/cycle counter at all, so we aren't
	 * going to bother with _bitvec[1] fields.
	 */
	for (this_inst = GetFirstInst(b); this_inst != NULL;
			this_inst = GetNextInst(this_inst))
	    {
	    InstRegUsageVec	this_insn_vec;

	    this_PC = InstPC (this_inst);	/* For debug purposes */
	    GetInstRegUsage(this_inst, &this_insn_vec);

		/*
		 * Compute the vector of registers which are used before they
		 * are set.  Or, registers which are used, but have not been set.
		 */

	    param_vec |= (this_insn_vec.ureg_bitvec[0] & ~this_proc_vec.dreg_bitvec[0]);

		/*
		 * Update the set of all set registers
		 */
	    this_proc_vec.dreg_bitvec[0] |= this_insn_vec.dreg_bitvec[0];
	    }	/* Each inst in the block */
	}	/* Each block in the proc */

	/*
	 * Each set bit in param_vec represents one of the following:
	 *	 a parameter
	 *	code with explicit knowledge of its callers register set
	 *	a bug in the code, compiler generated or otherwise
	 * Assume the first scenario.
	 */
    if ((param_vec & PARAM_A0_MASK_) != 0)
	{
	AddCallProc (p, ProcBefore, "Glissade_Int", ParamSequenceNum, REG_16);
	A0_SeqNum = ParamSequenceNum;
	ParamSequenceNum++;
#ifdef DEBUG
	fprintf (Inst_Logfile, "A0 (%d) ", A0_SeqNum);
#else
#ifndef SUMMARY_ONLY
	fprintf (Inst_Logfile, "A0 ");
#endif
#endif
	}
    else
	A0_SeqNum = -1;

    if ((param_vec & PARAM_A1_MASK_) != 0)
	{
	AddCallProc (p, ProcBefore, "Glissade_Int", ParamSequenceNum, REG_17);
	A1_SeqNum = ParamSequenceNum;
	ParamSequenceNum++;
#ifdef DEBUG
	fprintf (Inst_Logfile, "A1 (%d) ", A1_SeqNum);
#else
#ifndef SUMMARY_ONLY
	fprintf (Inst_Logfile, "A1 ");
#endif
#endif
	}
    else
	A1_SeqNum = -1;

    if ((param_vec & PARAM_A2_MASK_) != 0)
	{
	AddCallProc (p, ProcBefore, "Glissade_Int", ParamSequenceNum, REG_18);
	A2_SeqNum = ParamSequenceNum;
	ParamSequenceNum++;
#ifdef DEBUG
	fprintf (Inst_Logfile, "A2 (%d) ", A2_SeqNum);
#else
#ifndef SUMMARY_ONLY
	fprintf (Inst_Logfile, "A2 ");
#endif
#endif
	}
    else
	A2_SeqNum = -1;

    if ((param_vec & PARAM_FP0_MASK_) != 0)
	{
	AddCallProc (p, ProcBefore, "Glissade_Float", ParamSequenceNum, FREG_16);
	FP0_SeqNum = ParamSequenceNum;
	ParamSequenceNum++;
#ifdef DEBUG
	fprintf (Inst_Logfile, "FP0 (%d) ", FP0_SeqNum);
#else
#ifndef SUMMARY_ONLY
	fprintf (Inst_Logfile, "FP0 ");
#endif
#endif
	}
    else
	FP0_SeqNum = -1;

    if ((param_vec & PARAM_FP1_MASK_) != 0)
	{
	AddCallProc (p, ProcBefore, "Glissade_Float", ParamSequenceNum, FREG_17);
	FP1_SeqNum = ParamSequenceNum;
	ParamSequenceNum++;
#ifdef DEBUG
	fprintf (Inst_Logfile, "FP1 (%d) ", FP1_SeqNum);
#else
#ifndef SUMMARY_ONLY
	fprintf (Inst_Logfile, "FP1 ");
#endif
#endif
	}
    else
	FP1_SeqNum = -1;

    if ((param_vec & PARAM_FP2_MASK_) != 0)
	{
	AddCallProc (p, ProcBefore, "Glissade_Float", ParamSequenceNum, FREG_18);
	FP2_SeqNum = ParamSequenceNum;
	ParamSequenceNum++;
#ifdef DEBUG
	fprintf (Inst_Logfile, "FP2 (%d) ", FP2_SeqNum);
#else
#ifndef SUMMARY_ONLY
	fprintf (Inst_Logfile, "FP2 ");
#endif
#endif
	}
    else
	FP2_SeqNum = -1;

#ifndef SUMMARY_ONLY
    fprintf (Inst_Logfile, "\n");
#endif
    AddCallProc (p, ProcBefore, "Procedure_Count", ProcSequenceNum);
	/*
	 * Build ProcTable up so we can pass its contents to the runtime
	 * _after_ we allocate space for it.
	 */
    if (ProcName(p) == NULL)
	strncat (&ProcTable[ProcSequenceNum].ProcName[0], "(null)", PROC_NAME_SIZE);
    else
	strncat (&ProcTable[ProcSequenceNum].ProcName[0], ProcName(p), PROC_NAME_SIZE);
    ProcTable[ProcSequenceNum].A0_SeqNum = A0_SeqNum;
    ProcTable[ProcSequenceNum].A1_SeqNum = A1_SeqNum;
    ProcTable[ProcSequenceNum].A2_SeqNum = A2_SeqNum;
    ProcTable[ProcSequenceNum].FP0_SeqNum = FP0_SeqNum;
    ProcTable[ProcSequenceNum].FP1_SeqNum = FP1_SeqNum;
    ProcTable[ProcSequenceNum].FP2_SeqNum = FP2_SeqNum;

    AddCallProgram (ProgramAfter, "Dump_Procedure_Info", ProcSequenceNum);
    ProcSequenceNum++;
    if (ProcSequenceNum > PROC_TABLE_TYPE_SIZE)
	{
	fprintf (stderr, "ProcSequenceNum (0x%lx) > PROC_TABLE_TYPE_SIZE (0x%lx)\n",
		ProcSequenceNum, PROC_TABLE_TYPE_SIZE);
	fprintf (stderr, "Adjust datatypes.h & rebuild\n");
	exit (1);
	}
    }	/* Each proc */
}

void InstrumentFini()
{
  int	i;
  long	netval;

  AddCallProto ("Dyn_ProcCountAllocate(long)");
  AddCallProgram (ProgramBefore, "Dyn_ProcCountAllocate", ProcSequenceNum);
  AddCallProto ("Dyn_ProcCountInit(long,char *,long,long,long,long,long,long)");
  for (i = 0; i < ProcSequenceNum; i++)
    {
    AddCallProgram (ProgramBefore, "Dyn_ProcCountInit",
			i, ProcTable[i].ProcName,
			ProcTable[i].A0_SeqNum, ProcTable[i].A1_SeqNum,
			ProcTable[i].A2_SeqNum, ProcTable[i].FP0_SeqNum,
			ProcTable[i].FP1_SeqNum, ProcTable[i].FP2_SeqNum);
    }

  AddCallProto ("Dyn_ParamCountAllocate(long)");
  AddCallProgram (ProgramBefore, "Dyn_ParamCountAllocate", ParamSequenceNum);
  AddCallProto ("DynamicFini()");
  AddCallProgram (ProgramAfter, "DynamicFini");

#ifndef SUMMARY_ONLY
  fprintf(Inst_Logfile, "A total of %d procedures found\n", ProcSequenceNum);
#endif
  fprintf (Inst_Logfile, "*** Dynamic data begin ***\n");
  fclose(Inst_Logfile);

}

